#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # Use of PYBIND11_MAKE_OPAQUE
# 
# [pybind11](https://pybind11.readthedocs.io/) automatically converts `std::vector` into python list. That's convenient but not necessarily efficient depending on how it is used after that. [PYBIND11_MAKE_OPAQUE](https://pybind11.readthedocs.io/en/stable/advanced/cast/stl.html#making-opaque-types) is used to create a [capsule](https://docs.python.org/3/c-api/capsule.html) to hold a pointer on the C++ object.

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# ## Two identical classes
# 
# Both of then creates random vectors equivalent to `std::vector<Tensor>`, and `Tensor ~  std::vector<float>`. The first one returns a capsule due `PYBIND11_MAKE_OPAQUE(std::vector<OneTensorFloat>)` inserted into the C++ code. The other one is returning a list.

# In[3]:


from cpyquickhelper.examples.vector_container_python import (
    RandomTensorVectorFloat, RandomTensorVectorFloat2)

rnd = RandomTensorVectorFloat(10, 10)
result = rnd.get_tensor_vector()
print(result)


# In[4]:


result_ref = rnd.get_tensor_vector_ref()
print(result_ref)


# In[5]:


rnd2 = RandomTensorVectorFloat2(10, 10)
result2 = rnd2.get_tensor_vector()
print(result2)


# In[6]:


result2_ref = rnd2.get_tensor_vector_ref()
print(result2_ref)


# In[7]:


get_ipython().run_line_magic('timeit', 'rnd.get_tensor_vector()')


# In[8]:


get_ipython().run_line_magic('timeit', 'rnd.get_tensor_vector_ref()')


# In[9]:


get_ipython().run_line_magic('timeit', 'rnd2.get_tensor_vector()')


# In[10]:


get_ipython().run_line_magic('timeit', 'rnd2.get_tensor_vector_ref()')


# ## Scenarii
# 
# Three possibilities:
# 
# * **list**: `std::vector<Tensor>` is converted into a list of copied Tensors
# * **capsule**: `std::vector<Tensor>` is converted into a capsule on a copied `std::vector<Tensor>`, the capsule still holds the pointer and is responsible to the deletion.
# * **ref**: `std::vector<Tensor>` is just return as a pointer. The cost of getting the pointer does not depend on the content size. It is somehow the low limit.

# ## Plots

# In[11]:


import itertools
from cpyquickhelper.numbers.speed_measure import measure_time
from tqdm import tqdm
import pandas

data = []
sizes = [1, 2, 5, 10, 20, 50, 100, 200, 500, 1000, 5000, 10000]
sizes = list(itertools.product(sizes, sizes))
for i, j in tqdm(sizes):
    if j >= 1000:
        if i > 1000:
            continue
    if i * j >= 1e6:
        repeat, number = 3, 3
    else:
        repeat, number = 10, 10
    rnd = RandomTensorVectorFloat(i, j)
    obs = measure_time(lambda: rnd.get_tensor_vector(), repeat=repeat, number=number, div_by_number=True)
    obs['name'] = 'capsule'
    obs['n_vectors'] = i
    obs['size'] = j
    data.append(obs)

    rnd2 = RandomTensorVectorFloat2(i, j)
    obs = measure_time(lambda: rnd2.get_tensor_vector(), repeat=repeat, number=number, div_by_number=True)
    obs['name'] = 'list'
    obs['n_vectors'] = i
    obs['size'] = j
    data.append(obs)

    obs = measure_time(lambda: rnd2.get_tensor_vector_ref(), repeat=repeat, number=number, div_by_number=True)
    obs['name'] = 'ref'
    obs['n_vectors'] = i
    obs['size'] = j
    data.append(obs)
    
df = pandas.DataFrame(data)
df.tail()


# In[12]:


piv = pandas.pivot_table(df, index=['n_vectors', 'size'], columns=['name'], values='average')
piv['ratio'] = piv['capsule'] / piv['list']
piv.tail()


# In[13]:


import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 2, figsize=(10, 4))
piv[['capsule', 'list', 'ref']].plot(logy=True, ax=ax[0], title='Capsule (OPAQUE) / list')
piv.sort_values('ratio', ascending=False)[['ratio']].plot(ax=ax[1], title='Ratio Capsule (OPAQUE) / list');


# In[14]:


flat = piv.reset_index(drop=False)[['n_vectors', 'size', 'ratio']]
flat_piv = flat.pivot('n_vectors', 'size', 'ratio')
flat_piv


# In[15]:


import numpy
import seaborn
seaborn.heatmap(numpy.minimum(flat_piv.values, 1), cmap="YlGnBu",
                xticklabels=list(flat_piv.index), yticklabels=list(flat_piv.columns));


# In[16]:





# In[17]: